/*
   Copyright (C) 1997-2001 Id Software, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

#include "tv_local.h"

#include "tv_relay_cull.h"

#include "tv_relay.h"

//============
//TV_Relay_FatPVS
//
//The client will interpolate the view position,
//so we can't use a single PVS point
//===========
void TV_Relay_FatPVS( relay_t *relay, vec3_t org )
{
	int leafs[128];
	int i, j, count;
	int longs;
	qbyte *src;
	vec3_t mins, maxs;

	for( i = 0; i < 3; i++ )
	{
		mins[i] = org[i] - 8;
		maxs[i] = org[i] + 8;
	}

	count = CM_BoxLeafnums( relay->cms, mins, maxs, leafs, sizeof( leafs )/sizeof( int ), NULL );
	if( count < 1 )
		Com_Error( ERR_FATAL, "TV_Relay_FatPVS: count < 1" );

	longs = CM_ClusterSize( relay->cms ) / sizeof( int );

	// convert leafs to clusters
	for( i = 0; i < count; i++ )
		leafs[i] = CM_LeafCluster( relay->cms, leafs[i] );

	memcpy( relay->fatpvs, CM_ClusterPVS( relay->cms, leafs[0] ), CM_ClusterSize( relay->cms ) );

	// or in all the other leaf bits
	for( i = 1; i < count; i++ )
	{
		for( j = 0; j < i; j++ )
			if( leafs[i] == leafs[j] )
				break;
		if( j != i )
			continue; // already have the cluster we want
		src = CM_ClusterPVS( relay->cms, leafs[i] );
		for( j = 0; j < longs; j++ )
			( (int *)relay->fatpvs )[j] |= ( (int *)src )[j];
	}
}

//============
//TV_Relay_FatPHS
//============
void TV_Relay_FatPHS( relay_t *relay, int cluster )
{
	memcpy( relay->fatphs, CM_ClusterPHS( relay->cms, cluster ), CM_ClusterSize( relay->cms ) );
}

//============
//TV_Relay_MergePVS
//
//Portal entities add a second PVS at origin2 to fatpvs
//===========
void TV_Relay_MergePVS( relay_t *relay, vec3_t org )
{
	int leafs[128];
	int i, j, count;
	int longs;
	qbyte *src;
	vec3_t mins, maxs;

	for( i = 0; i < 3; i++ )
	{
		mins[i] = org[i] - 1;
		maxs[i] = org[i] + 1;
	}

	count = CM_BoxLeafnums( relay->cms, mins, maxs, leafs, sizeof( leafs )/sizeof( int ), NULL );
	if( count < 1 )
		Com_Error( ERR_FATAL, "TV_Relay_MergePVS: count < 1" );

	longs = CM_ClusterSize( relay->cms ) / sizeof( int );

	// convert leafs to clusters
	for( i = 0; i < count; i++ )
		leafs[i] = CM_LeafCluster( relay->cms, leafs[i] );

	// or in all the other leaf bits
	for( i = 0; i < count; i++ )
	{
		for( j = 0; j < i; j++ )
			if( leafs[i] == leafs[j] )
				break;
		if( j != i )
			continue; // already have the cluster we want
		src = CM_ClusterPVS( relay->cms, leafs[i] );
		for( j = 0; j < longs; j++ )
			( (int *)relay->fatpvs )[j] |= ( (int *)src )[j];
	}
}

//============
//TV_Relay_MergePHS
//===========
void TV_Relay_MergePHS( relay_t *relay, int cluster )
{
	int i, longs;
	qbyte *src;

	longs = CM_ClusterSize( relay->cms ) / sizeof( int );

	// or in all the other leaf bits
	src = CM_ClusterPHS( relay->cms, cluster );
	for( i = 0; i < longs; i++ )
		( (int *)relay->fatphs )[i] |= ( (int *)src )[i];
}

//============
//TV_Relay_PVSCullEntity
//===========
qboolean TV_Relay_PVSCullEntity( relay_t *relay, edict_t *ent )
{
	int i, l;

	// too many leafs for individual check, go by headnode
	if( ent->r.num_clusters == -1 )
	{
		if( !CM_HeadnodeVisible( relay->cms, ent->r.headnode, relay->fatpvs ) )
			return qtrue;
		return qfalse;
	}

	// check individual leafs
	for( i = 0; i < ent->r.num_clusters; i++ )
	{
		l = ent->r.clusternums[i];
		if( relay->fatpvs[l >> 3] & ( 1 << ( l&7 ) ) )
		{
			return qfalse;
		}
	}

	return qtrue; // not visible
}

//============
//TV_Relay_PHSCullEntity
//===========
qboolean TV_Relay_PHSCullEntity( relay_t *relay, edict_t *ent )
{
	int l;

	// too many leafs for individual check, go by headnode
	if( ent->r.num_clusters == -1 )
	{
		if( !CM_HeadnodeVisible( relay->cms, ent->r.headnode, relay->fatphs ) )
			return qtrue;
		return qfalse;
	}

	if( ent->r.num_clusters )
	{
		l = ent->r.clusternums[0];
		if( ( relay->fatphs[l >> 3] & ( 1 << ( l&7 ) ) ) )
			return qfalse;
	}

	return qtrue; // not audible
}

//============
//TV_Relay_SnapCullEntity
//===========
qboolean TV_Relay_SnapCullEntity( relay_t *relay, edict_t *ent, edict_t *clent, client_frame_t *frame,
                                  vec3_t vieworg, int clientarea )
{
	// filters: this entity has been disabled for communication
	if( ent->r.svflags & SVF_NOCLIENT )
		return qtrue;

	// send all entities
	if( frame->allentities )
		return qfalse;

	// filters: transmit only to clients in the same team as this entity
	// wsw: imp: moved if because broadcasting is less important than team specifics
	if( ( ent->r.svflags & SVF_ONLYTEAM ) && ( clent && ent->s.team != clent->s.team ) )
		return qtrue;

	if( ent->r.svflags & SVF_BROADCAST )  // send to everyone
		return qfalse;

	// SPECIAL CASE FOR LINEAR PROJECTILES:
	// since there is no game module, and linear projectiles don't transmit
	// their actual origin, we have to extrapolate them for culling.
	if( ent->s.linearProjectile )
	{
		return qfalse;
		/*
		// linear projectiles do not have a real origin, it is extrapolated from static data
		if( relay->serverTime > ( ent->s.linearProjectileTimeStamp + relay->delay ) )
		{
			float flyTime = (float)( relay->serverTime - ( ent->s.linearProjectileTimeStamp + relay->delay ) ) * 0.001f;
			VectorMA( ent->s.origin2, flyTime, ent->s.linearProjectileVelocity, ent->s.origin );
		}
		else
			VectorCopy( ent->s.origin2, ent->s.origin );
			*/
	}

	// if the entity is only a sound
	if( !ent->s.modelindex && !ent->s.events[0] && !ent->s.light && !ent->s.effects && ent->s.sound )
	{
		if( DistanceFast( ent->s.origin, vieworg ) > 400 )
		{   // it would be attenuated
			return qtrue;
		}
	}

	// jal: is this the same as CM_AreasConnected??? (this one includes the portal, client area doesn't)
	if( !( frame->areabits[ent->r.areanum>>3] & ( 1<<( ent->r.areanum&7 ) ) ) )
	{
		// doors can legally straddle two areas, so we may need to check another one
		if( !ent->r.areanum2 || !( frame->areabits[ent->r.areanum2>>3] & ( 1<<( ent->r.areanum2&7 ) ) ) )
			return qtrue; // blocked by a door
	}

	if( ent->r.svflags & SVF_TRANSMITORIGIN2 )
	{                                        // jal: they don't move. No need to use fatPVS (I think)
		// cull by PHS
		if( TV_Relay_PHSCullEntity( relay, ent ) )
			return qtrue;

	}
	else
	{
		// cull by PVS
		if( TV_Relay_PVSCullEntity( relay, ent ) )
			return qtrue;
	}

	return qfalse; // not culled
}
